(define-record-type <CmdSeq> (CmdSeq commands) CmdSeq?
  (commands %cmdseq-commands %set-cmdseq-commands!) ;; list
  )

(define (cmdseq-add! cmdseq command)
  (%set-cmdseq-commands! cmdseq (cons command (%cmdseq-commands cmdseq))))

(define (cmdseq->list cmdseq)
  (reverse (%cmdseq-commands cmdseq)))


(define-record-type <CModule> (CModule toplevel initializer gensymidx) CModule?
  (toplevel cmodule-toplevel) ;; CmdSeq
  (initializer cmodule-initializer) ;; CFunction
  (gensymidx cmodule-gensymidx set-cmodule-gensymidx!) ;; fixnum
  )

(define-record-type <CFunction> (CFunction signature body gensymidx captures) CFunction?
  (signature cfunction-signature)
  (body cfunction-body)
  (gensymidx cfunction-gensymidx set-cfunction-gensymidx!)
  (captures cfunction-captures set-cfunction-captures!))

(define-record-type <CBlock> (CBlock pretext cmdseq) CBlock?
  (pretext cblock-pretext)
  (cmdseq cblock-cmdseq))

(define-record-type <CStatement> (CStatement text) CStatement?
  (text cstatement-text))

(define-record-type <CEnv> (CEnv bindings parent) CEnv?
  (bindings cenv-bindings) ;; dict (Binding -> Accessor)
  (parent cenv-parent) ;; CEnv or #f
  )

(define (cenv-ref cenv binding)
  (and cenv
       (or (dict-ref (cenv-bindings cenv) binding #f)
           (cenv-ref (cenv-parent cenv) binding))))

(define (merge-cenvs cenv1 cenv2)
  (cond ((not cenv1) cenv2)
        (else (CEnv (cenv-bindings cenv1)
                    (merge-cenvs (cenv-parent cenv1) cenv2)))))


(define-record-type <CContext>
  (CContext cmodule cfunction cblock
            global-cenv local-cenv captured-cenv)
  CContext?
  ;;
  (cmodule ccontext-cmodule)
  (cfunction ccontext-cfunction)
  (cblock ccontext-cblock)
  (global-cenv ccontext-global-cenv)
  (local-cenv ccontext-local-cenv)
  (captured-cenv ccontext-captured-cenv))


(define (text . components)
  (with-output-to-buffer
    (lambda (output)
      (let loop ((components components))
        (cond ((null? components) #f)
              (else (output (car components))
                    (loop (cdr components))))))))


(define (gen-global-cenv mod)
  (let ((cbindings (make-dict)))
    (dict-for-each
     (lambda (name binding)
       (dict-set! cbindings
                  binding
                  (text "helTop_" (c-escape-name (symbol->string (binding-name binding))))))
     (module-bindings mod))
    (CEnv cbindings #f)))

(define (make-context cmod cfun global-cenv captured-cenv)
  (CContext cmod cfun (cfunction-body cfun)
            global-cenv (CEnv (make-dict) #f) captured-cenv))


(define (toplevel->c mod top)
  (let* ((ctx (make-global-context mod)))
    (c-statement! ctx (text "Val callinfo = VAL_FALSE;"))
    (for-each (lambda (il) (c-form ctx il)) top)
    ctx))

(define (make-global-context mod)
  (let* ((global-cenv (gen-global-cenv mod))
         (initializer (CFunction
                       "__attribute__((constructor)) static void init_module(void)"
                       (CBlock #f (CmdSeq '()))
                       0
                       '()))
         (cmod (CModule (CmdSeq '()) initializer 0)))
    (make-context cmod initializer global-cenv #f)))

;; The translation functions take an expression and return a 'text' which can be
;; used to access the computed expression value. They also mutate the context to
;; append the necessary statements to compute the expression.
;;
;; The returned expression should be free of side-effects; for instance, the
;; translation of a function call must return a variable containing the result
;; of the call, not a function call expression. The returned value may be simply
;; discarded if the expression was called in a context which does not use the
;; value (i.e., just for side-effects).


(define (c-form ctx il)
  (cond
   ((Const? il)
    (cond ((eq? 'fixnum (const-type il)) (text "INT_TO_FIXNUM(" (const-value il) ")"))
          ((eq? 'string (const-type il)) (c-string-literal ctx il))
          ((eq? 'boolean (const-type il)) (text (if (const-value il) "VAL_TRUE" "VAL_FALSE")))
          (else (error "c-form: Unknown constant type"))))
   ((Var? il) (ctx-fetch ctx (var-binding il)))
   ((Seq? il) (begin (c-form ctx (seq-left il))
                     (c-form ctx (seq-right il))))
   ((If? il) (c-if ctx il))
   ((Lambda? il) (c-lambda ctx il))
   ((Funcall? il) (c-funcall ctx il))
   ((CValue? il) (c-cvalue ctx il))
   ((Define? il) (c-define ctx il))
   (else (error "c-form got lost"))))

(define (c-if ctx il)
  (let* ((result-var (make-temp-var ctx "if_result"))
           (test-value (c-form ctx (if-test il))))
      (c-block! ctx (text "if (VAL_IS_TRUE(" test-value "))")
        (lambda (ctx)
          (c-statement! ctx (text result-var " = " (c-form ctx (if-then il)) ";"))))
      (c-block! ctx (text "else")
        (lambda (ctx)
          (c-statement! ctx (text result-var " = " (c-form ctx (if-else il)) ";"))))
      result-var))

(define (c-lambda ctx il)
  (let* ((cfun-name (cmodule-gensym! (ccontext-cmodule ctx) "clofun"))
         (cfun (CFunction (text "Val " cfun-name "(Val callinfo[])")
                          (CBlock #f (CmdSeq '()))
                          0
                          '()))
         (subctx (CContext (ccontext-cmodule ctx)
                           cfun
                           (cfunction-body cfun)
                           (ccontext-global-cenv ctx)
                           (make-params-cenv (lambda-params il))
                           (merge-cenvs (ccontext-captured-cenv ctx)
                                        (ccontext-local-cenv ctx))))
         (return-value (c-form subctx (lambda-body il))))
    (c-statement! subctx (text "return " return-value ";"))
    (c-toplevel! ctx cfun)
    (c-alloc-closure ctx cfun-name cfun)
    ))

(define (c-alloc-closure ctx cfun-name cfun)
  (let* ((captures (cfunction-captures cfun))
         (closure-var (make-temp-var ctx "clo")))
    (c-statement! ctx (text closure-var " = ALLOC_WORD_BLOB("
                            (+ closure-firstvar-idx (length captures))
                            ");"))
    (c-statement! ctx (text "WORD_REF(" closure-var ", " closure-tag-idx ") = XTAG_CLOSURE;"))
    (c-statement! ctx (text "WORD_REF(" closure-var ", " closure-function-idx ") = (Val) " cfun-name ";"))
    (c-statement! ctx (text "WORD_REF(" closure-var ", " closure-numvars-idx ") = INT_TO_FIXNUM(" (length captures) ");"))
    (let loop ((i closure-firstvar-idx)
               (captures (reverse captures)))
      ;; TODO: tag, fun, len
      (when (not (null? captures))
        (c-statement! ctx (text "WORD_REF(" closure-var ", " i ") = "
                                (ctx-fetch ctx (car (car captures))) ";"))
        (loop (+ i 1) (cdr captures))))
    closure-var))

(define (c-funcall ctx il)
  (let ((result (make-temp-var ctx "fun_result"))
        (subcallinfo (cfunction-gensym! (ccontext-cfunction ctx) "subcallinfo")))
    (c-statement! ctx (text "Val " subcallinfo "[" (+ (length (funcall-args il))
                                                  callinfo-firstarg-idx) "];"))
    (c-statement! ctx (text subcallinfo "[" callinfo-caller-idx "] = (Val) callinfo;"))
    (c-statement! ctx (text subcallinfo "[" callinfo-callee-idx "] = " (c-form ctx (funcall-fun il)) ";"))
    (c-statement! ctx (text subcallinfo "[" callinfo-numargs-idx "] = INT_TO_FIXNUM(" (length (funcall-args il)) ");"))
    (let loop ((i callinfo-firstarg-idx)
               (args (funcall-args il)))
      (when (not (null? args))
        (c-statement! ctx (text subcallinfo "[" i "] = " (c-form ctx (car args)) ";"))
        (loop (+ i 1) (cdr args))))
    (c-statement! ctx (text result " = ((Fun) WORD_REF(" subcallinfo "[" callinfo-callee-idx "], " closure-function-idx "))(" subcallinfo ");"))
    result))







(define (make-params-cenv params)
  (let ((dict (make-dict)))
    (let loop ((params params) (i 0))
      (cond ((null? params) (CEnv dict #f))
            (else (dict-set! dict (var-binding (car params))
                             (text "callinfo[" (+ i callinfo-firstarg-idx) "]"))
                  (loop (cdr params) (+ i 1)))))))

(define (c-block! ctx pretext proc)
  (let ((block (CBlock pretext (CmdSeq '()))))
    (proc (CContext (ccontext-cmodule ctx)
                    (ccontext-cfunction ctx)
                    block
                    (ccontext-global-cenv ctx)
                    (ccontext-local-cenv ctx)
                    (ccontext-captured-cenv ctx)))
    (cmdseq-add! (cblock-cmdseq (ccontext-cblock ctx)) block)))

(define (c-statement! ctx text)
  (cmdseq-add! (cblock-cmdseq (ccontext-cblock ctx)) (CStatement text)))

(define (c-toplevel! ctx text)
  (cmdseq-add! (cmodule-toplevel (ccontext-cmodule ctx)) text))

(define (cfunction-gensym! cfun mnemonic)
  (let ((index (cfunction-gensymidx cfun)))
    (set-cfunction-gensymidx! cfun (+ index 1))
    (text "tmp_" mnemonic index)))

(define (cmodule-gensym! cmod mnemonic)
  (let ((index (cmodule-gensymidx cmod)))
    (set-cmodule-gensymidx! cmod (+ index 1))
    (text "helTmp_" mnemonic index)))


(define (make-temp-var ctx mnemonic)
  (let ((varname (cfunction-gensym! (ccontext-cfunction ctx) mnemonic)))
    (c-statement! ctx (text "Val " varname " = VAL_UNSPECIFIED;"))
    varname))


(define callinfo-caller-idx 0)
(define callinfo-callee-idx 1)
(define callinfo-numargs-idx 2)
(define callinfo-firstarg-idx 3)

(define closure-tag-idx 0)
(define closure-function-idx 1)
(define closure-numvars-idx 2)
(define closure-firstvar-idx 3)

(define (ctx-fetch ctx binding)
  (cond
   ((cenv-ref (ccontext-local-cenv ctx) binding)
    => (lambda (binding) binding))
   ((cenv-ref (ccontext-captured-cenv ctx) binding)
    (cond ((assq binding (cfunction-captures (ccontext-cfunction ctx)))
           => cdr)
          (else
           (let ((accessor (text "WORD_REF(callinfo[CALLINFO_CLOSURE_IDX], "
                                 (+ (length (cfunction-captures (ccontext-cfunction ctx)))
                                    closure-firstvar-idx)
                                 ")")))
             (set-cfunction-captures! (ccontext-cfunction ctx)
                                      (cons (cons binding accessor)
                                            (cfunction-captures (ccontext-cfunction ctx))))
             accessor))))
   ((cenv-ref (ccontext-global-cenv ctx) binding)
    => (lambda (binding) binding))
   (else (newline) (display binding) (newline) ;; debug
         (error "Unbound cenv variable (shouldn't happen!)" binding ctx))))


(define (c-string-literal ctx il)
  (let* ((strdata (cmodule-gensym! (ccontext-cmodule ctx) "strdata"))
         (strval (cmodule-gensym! (ccontext-cmodule ctx) "strval"))
         (string (c-escape-string (const-value il))))
    (c-toplevel! ctx (CStatement (text "static Val " strdata "[] = { XTAG_STRING, INT_TO_FIXNUM(" (string-length (const-value il)) "), \"" string "\" };")))
    (c-toplevel! ctx (CStatement (text "static Val " strval " = TAG_PTR(" strdata ");")))
    strval))

(define (c-escape-string str)
  (with-output-to-buffer
    (lambda (output)
      (let ((len (string-length str)))
        (let loop ((i 0))
          (when (< i len)
            (let ((ch (string-ref str i)))
              (cond ((eqv? ch #\newline) (output "\\n"))
                    ((eqv? ch #\") (output "\\\""))
                    (else (output ch)))
              (loop (+ i 1)))))))))


(define (c-escape-name str)
  (with-output-to-buffer
    (lambda (output)
      (let ((len (string-length str)))
        (let loop ((i 0))
          (when (< i len)
            (let ((ch (string-ref str i)))
              (cond ((eqv? ch #\-) (output "_"))
                    ((or (and (char<=? #\a ch) (char<=? ch #\z))
                         (and (char<=? #\A ch) (char<=? ch #\Z))
                         (and (char<=? #\0 ch) (char<=? ch #\9)))
                     (output ch))
                    (else
                     (output "$")
                     (output (char->integer ch))
                     (output "$")))
              (loop (+ i 1)))))))))

(define (c-cvalue ctx il)
  (let ((items (map (lambda (item)
                      (c-form ctx item))
                    (cvalue-items il))))
    (apply text (replace-string-template (cvalue-template il) items))))


(define (c-define ctx il)
  ;; TODO: currently we assume top-level define only.
  (let ((name (ctx-fetch ctx (var-binding (define-name il)))))
    (c-toplevel! ctx (CStatement (text "Val " name " = VAL_UNSPECIFIED;")))
    (c-statement! ctx (text name " = " (c-form ctx (define-value il)) ";"))
    (text "VAL_UNSPECIFIED")))

;;; Printing the actual C code.

(define (emit-c level thing)
  (cond
   ((CContext? thing)
    (emit-c level (ccontext-cmodule thing)))

   ((CModule? thing)
    (emit-c level (cmodule-toplevel thing))
    (emit-c level (cmodule-initializer thing)))

   ((CmdSeq? thing)
    (for-each (lambda (form)
                (emit-c level form))
              (cmdseq->list thing)))

   ((CFunction? thing)
    (emit-text (cfunction-signature thing))
    (emit-text " ")
    (emit-c level (cfunction-body thing)))

   ((CBlock? thing)
    (emit-indent level)
    (when (cblock-pretext thing)
      (emit-text (cblock-pretext thing))
      (emit-text " "))
    (emit-text "{\n")
    (emit-c (+ 1 level) (cblock-cmdseq thing))
    (emit-indent level)
    (emit-text "}\n"))

   ((CStatement? thing)
    (emit-indent level)
    (emit-text (cstatement-text thing))
    (emit-text "\n"))

   ;; ((string? thing)
   ;;  (display thing))
   (else (error 'emit-c "Don't know how to emit this" thing))))

(define (emit-text text)
  (cond ((string? text) (display text))
        (else (error 'emit-text "Expecting string" text))))

(define (emit-indent level)
  (cond ((<= level 0) #f)
        (else (emit-text "    ")
              (emit-indent (- level 1)))))
