(define-record-type loc
  (make-loc line column)
  loc?
  ;;
  (line loc-line set-loc-line!)
  (column loc-column set-loc-column!))

(define (loc-advance-cols! loc cols)
  (set-loc-column! loc (+ (loc-column loc) cols)))

(define (loc-advance-line! loc)
  (set-loc-column! loc 1)
  (set-loc-line! loc (+ 1 (loc-line loc))))

(define (copy-loc loc)
  (make-loc (loc-line loc) (loc-column loc)))


(define-record-type reader
  (make-reader port filename loc)
  reader?
  ;;
  (port reader-port)
  (filename reader-filename)
  (loc reader-loc set-reader-loc!))

(define (reader-peek-char reader)
  (peek-char (reader-port reader)))

(define (reader-read-char reader)
  (let ((char (read-char (reader-port reader))))
    (if (eqv? char #\newline)
        (loc-advance-line! (reader-loc reader))
      (loc-advance-cols! (reader-loc reader) 1))
    char))

(define (reader-error loc msg . args)
  (display "Reader error at position ")
  (write loc)
  (display ": ")
  (display msg)
  (newline)
  (for-each (lambda (arg)
              (write arg)
              (newline))
            args)
  (error "Reader error"))


(define (whitespace? char)
  (or (eqv? char #\space)
      (eqv? char #\newline)
      (eqv? char #\tab)))

(define (delimiter? char)
  (or (whitespace? char)
      (eqv? char #\()
      (eqv? char #\))))

(define (digit? char)
  (and (char>=? char #\0)
       (char<=? char #\9)))

(define (skip-line! reader)
  (let ((char (reader-read-char reader)))
    (or (eof-object? char)
        (eqv? char #\newline)
        (skip-line! reader))))

(define (skip-whitespace! reader)
  (let ((char (reader-peek-char reader)))
    (cond ((whitespace? char)
           (reader-read-char reader)
           (skip-whitespace! reader))
          ((eqv? char #\;)
           (skip-line! reader)
           (skip-whitespace! reader))
          (else #t))))


(define (reader-read reader)
  (skip-whitespace! reader)
  (let* ((loc (copy-loc (reader-loc reader)))
         (char (reader-peek-char reader)))
    (cond
     ((eof-object? char) char)
     ((eqv? char #\()
      (reader-read-delimited-list loc #\( reader #\)))
     ((eqv? char #\#) (reader-read-hash loc reader))
     ((eqv? char #\") (reader-read-string loc reader))
     (else (reader-read-token loc reader)))))

(define (reader-read-hash loc reader)
  (or (eqv? (reader-read-char reader) #\#)
      (reader-error loc "Expecting #"))
  (let ((char (reader-read-char reader)))
    (cond
     ((eqv? char #\\) (reader-read-char-literal loc reader))
     ((or (eqv? char #\t) (eqv? char #\f))
      (make-token loc 'boolean (eqv? char #\t)))
     (else (reader-error loc "Unsupported hash syntax" char)))))

(define (reader-read-char-literal loc reader)
  (let ((char (reader-peek-char reader)))
    (if (delimiter? char)
        (make-token loc 'character (reader-read-char reader)))
      (let ((text (reader-read-token-text reader)))
        (make-token
         loc
         'character
         (cond ((= (string-length text) 1)
                (string-ref text 0))
               ((equal? text "space") #\space)
               ((equal? text "tab") #\tab)
               ((equal? text "newline") #\newline)
               (else (reader-error loc "Unknown character name ~a" text)))))))



(define-record-type token
  (make-token loc type value)
  token?
  (loc token-loc)
  (type token-type)
  (value token-value))

(define (reader-read-token-text reader)
  (let loop ((chars '()))
    (let ((char (reader-peek-char reader)))
      (cond ((or (delimiter? char) (eof-object? char))
             (list->string (reverse chars)))
            (else (loop (cons (reader-read-char reader) chars)))))))

(define (text-numeric? text)
  (let loop ((i (- (string-length text) 1)))
    (cond ((< i 0) #t)
          ((not (digit? (string-ref text i))) #f)
          (else (loop (- i 1))))))

(define (reader-read-token loc reader)
  (let ((text (reader-read-token-text reader)))
    (if (zero? (string-length text))
        (reader-error loc "Expecting token, found nothing")
      (if (text-numeric? text)
          (make-token loc 'number (string->number text))
        (make-token loc 'symbol (string->symbol text))))))

(define (reader-read-delimited-list loc open-char reader close-char)
  (or (eqv? (reader-read-char reader) open-char)
      (reader-error loc "Expecting list open-char"))
  (let loop ((items '()))
    (skip-whitespace! reader)
    (let ((char (reader-peek-char reader)))
      (cond ((eof-object? char)
             (reader-error loc "EOF while reading list, started at ~a" loc))
            ((eqv? char close-char)
             (reader-read-char reader)
             (make-token loc 'list (reverse items)))
            (else
             (let ((item (reader-read reader)))
               ;;(write item) (newline)
               (loop (cons item items))))))))

(define (reader-read-string loc reader)
  (or (eqv? (reader-read-char reader) #\")
      (reader-error loc "Expecting double-quote"))
  (let loop ((chars '()))
    (let ((char (reader-read-char reader)))
      (cond ((eqv? char #\")
             (make-token loc 'string (list->string (reverse chars))))
            ((eof-object? char)
             (reader-error loc "EOF while reading string"))
            ((eqv? char #\\)
             (let* ((next-char (reader-read-char reader))
                    (interpretation (if (eqv? next-char #\n)
                                        #\newline
                                      next-char)))
               ;; TODO: Handle backslash escapes properly.
               (loop (cons interpretation chars))))
            (else (loop (cons char chars)))))))


(define (reader-read-all reader)
  (let loop ((items '()))
    (let ((item (reader-read reader)))
      (cond ((eof-object? item)
             (reverse items))
            (else
             ;;(write item) (newline)
             (loop (cons item items)))))))

(define (reader-read-file filename)
  (call-with-input-file filename
    (lambda (port)
      (let ((reader (make-reader port filename (make-loc 1 1))))
        (reader-read-all reader)))))

(define (hel-read-from-string str)
  (call-with-input-string str
    (lambda (port)
      (let ((reader (make-reader port "stdin" (make-loc 1 1))))
        (reader-read reader)))))
