(define-record-type <Module>
  (Module bindings)
  Module?
  ;;
  (bindings module-bindings set-module-bindings!) ;; dict (name -> binding)
  )

(define (module-ref mod name)
  (dict-ref (module-bindings mod) name #f))

(define (module-set! mod name binding)
  (dict-set! (module-bindings mod) name binding))

(define (module-add mod name binding)
  (dict-set! (module-bindings mod) name binding))

(define-record-type <Binding>
  (Binding name module defined? kind props)
  Binding?
  ;;
  (name binding-name set-binding-name!) ;; symbol
  (module binding-module set-binding-module!) ;; module or #f if local
  (defined? binding-defined? set-binding-defined?!) ;; boolean
  (kind binding-kind set-binding-kind!) ;; value | syntax | special
  (props binding-props set-binding-props!) ;; dict
  )

(define-record-type <Env>
  (Env bindings parent)
  Env?
  ;;
  (bindings env-bindings set-env-bindings!) ;; dict (name -> binding)
  (parent env-parent set-env-parent!) ;; env or #f
  )

(define (env-ref env name)
  (and env
       (or (dict-ref (env-bindings env) name #f)
           (env-ref (env-parent env) name))))


;; Special forms.

(define *mod* (Module (make-dict)))
(define *env* #f) ;;(Env (make-dict) #f))

(define (register-special-form! name expander)
  (let ((binding (Binding name *mod* #t 'special (make-dict))))
    (dict-set! (binding-props binding) 'special-expander expander)
    (module-set! *mod* name binding)))


(define (if->treeil mod env fullform _operator operands)
  (let* ((test (car operands))
         (then-expr (cadr operands))
         (else-expr (caddr operands))
         (trailing (cdddr operands)))
    (if (not (null? trailing))
        (src-error fullform "Extra forms in 'if'")
      (If (form->treeil mod env test)
          (form->treeil mod env then-expr)
          (form->treeil mod env else-expr)))))

(register-special-form! 'if if->treeil)


(define (from-token type token)
  (if (eq? (token-type token) type)
      (token-value token)
    (error "Unexpected token type:" type token)))

(define (lambda->treeil mod env fullform _operator operands)
  (let* ((params (map (lambda (token) (from-token 'symbol token))
                      (from-token 'list (car operands))))
         (body (cdr operands))
         (bindings (map (lambda (param)
                          (Binding #;name param
                                   #;mod #f
                                   #;defined #t
                                   #;kind 'value
                                   #;props (make-dict)))
                        params))
         (new-env (Env (alist->dict (map cons params bindings)) env)))
    (Lambda (map Var bindings)
            (begin->treeil mod new-env fullform _operator body))))

(register-special-form! 'lambda lambda->treeil)


(define (begin->treeil mod env fullform _operator operands)
  (cond ((null? operands) (Unspecified))
        ((null? (cdr operands)) (form->treeil mod env (car operands)))
        (else (Seq (form->treeil mod env (car operands))
                   (begin->treeil mod env fullform _operator (cdr operands))))))

(register-special-form! 'begin begin->treeil)


(define (define->treeil mod env fullform _operator operands)
  (let ((name (car operands))
        (value (cadr operands))
        (trailing (cddr operands)))
    (cond ((not (null? trailing))
           (src-error fullform "Extra forms in 'define'"))
          (env (src-error fullform "Non-toplevel 'define' currently not supported"))
          (else
           (let ((binding (get-binding mod env (token-value name))))
             (if (binding-defined? binding)
                 (src-warn fullform "'define' of already-defined variable" name)
               #f)
             (set-binding-defined?! binding #t)
             (Define (Var binding) (form->treeil mod env value)))))))

(register-special-form! 'define define->treeil)


(define (cvalue->il mod env fullform _operator operands)
  (let ((template (car operands))
        (items (cdr operands)))
    (CValue (from-token 'string template)
            (map (lambda (item)
                   (form->treeil mod env item))
                 items))))

(register-special-form! '%c-value cvalue->il)

;; TreeIL forms:

(define-record-type <Const> (Const type value) Const?
  (type const-type)
  (value const-value))

(define-record-type <Var> (Var binding) Var?
  (binding var-binding))

(define-record-type <Funcall> (Funcall fun args) Funcall?
  (fun funcall-fun)
  (args funcall-args))

(define-record-type <If> (If test then else) If?
  (test if-test)
  (then if-then)
  (else if-else))

(define-record-type <Seq> (Seq left right) Seq?
  (left seq-left)
  (right seq-right))

(define-record-type <Lambda> (Lambda params body) Lambda?
  (params lambda-params)
  (body lambda-body))

(define-record-type <Define> (Define name value) Define?
  (name define-name)
  (value define-value))

(define-record-type <Toplevel> (Toplevel forms) Toplevel?
  (forms toplevel-forms))

(define-record-type <CValue> (CValue template items) CValue?
  (template cvalue-template)
  (items cvalue-items))

(define (Unspecified) (Const 'special 'unspecified))

;; ----

(define (get-binding mod env name)
  (or (env-ref env name)
      (module-ref mod name)
      (let ((binding (Binding name mod #f 'value (make-dict))))
        (module-set! mod name binding)
        binding)))

(define (form->treeil mod env form-token)
  (let ((type (token-type form-token))
        (value (token-value form-token)))
    (cond
     ((eq? type 'symbol)
      ;; TODO: binding may be syntax!
      (Var (get-binding mod env value)))
     ((eq? type 'list)
      (if (null? value)
          (src-error form-token "Invalid form: unquoted empty list")
        (combination->treeil mod env form-token (car value) (cdr value))))
     ((eq? type 'number)
      (Const 'fixnum value))
     ((eq? type 'string)
      (Const 'string value))
     ((eq? type 'boolean)
      (Const 'boolean value))
     (else
      (src-error form-token "Don't know how to compile that")))))

(define (combination->treeil mod env fullform operator operands)
  (cond
   ((eq? (token-type operator) 'symbol)
    (let ((operator-binding (get-binding mod env (token-value operator))))
      (cond ((eq? (binding-kind operator-binding) 'syntax)
             (src-error operator "TODO: Handle syntax forms"))
            ((eq? (binding-kind operator-binding) 'value)
             (funcall->treeil mod env operator operands))
            ((eq? (binding-kind operator-binding) 'special)
             (let ((expander (dict-ref (binding-props operator-binding) 'special-expander #f)))
               (expander mod env fullform operator operands)))
            (else (src-error operator "Unknown binding kind" operator-binding)))))
   ((eq? (token-type operator) 'list)
    (funcall->treeil mod env operator operands))
   (else
    (src-error fullform "Invalid form at operator position" operator))))

(define (funcall->treeil mod env operator operands)
  (Funcall (form->treeil mod env operator)
           (map (lambda (arg)
                  (form->treeil mod env arg))
                operands)))

(define (src-error token msg . args)
  (apply src-report token "Syntax ERROR: " msg args)
  (error "Syntax error"))

(define (src-warn token msg . args)
  (apply src-report token "Syntax WARNING: " msg args))

(define (src-report token prefix msg . args)
  (with-output-to-port (current-error-port)
    (lambda ()
      (display prefix)
      (display msg)
      (newline)
      (for-each (lambda (arg)
                  (write arg)
                  (newline))
                args)
      (write token)
      (newline))))


;;; Debugging utils.

(define (vector-map fn vec)
  (let* ((len (vector-length vec))
         (vec* (make-vector len)))
    (let loop ((i 0))
      (cond ((< i len)
             (vector-set! vec* i (fn (vector-ref vec i)))
             (loop (+ i 1)))
            (else vec*)))))


(define (treeil-basic il)
  (cond ((Binding? il) (vector '<Binding*> (binding-name il)))
        ((vector? il) (vector-map treeil-basic il))
        ((pair? il) (cons (treeil-basic (car il))
                          (treeil-basic (cdr il))))
        (else il)))

(define (treeil-from-string str)
  (form->treeil *mod* *env* (hel-read-from-string str)))
